/*
 * Copyright (c) 2015, 2019, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

#ifndef SHARE_GC_X_XPAGETABLE_HPP
#define SHARE_GC_X_XPAGETABLE_HPP

#include "gc/x/xGranuleMap.hpp"
#include "memory/allocation.hpp"

class VMStructs;
class XPage;
class XPageTableIterator;

class XPageTable {
  friend class ::VMStructs;
  friend class XPageTableIterator;

private:
  XGranuleMap<XPage*> _map;

public:
  XPageTable();

  XPage* get(uintptr_t addr) const;

  void insert(XPage* page);
  void remove(XPage* page);
};

class XPageTableIterator : public StackObj {
private:
  XGranuleMapIterator<XPage*> _iter;
  XPage*                      _prev;

public:
  XPageTableIterator(const XPageTable* page_table);

  bool next(XPage** page);
};

#endif // SHARE_GC_X_XPAGETABLE_HPP
