/*
 * Copyright (c) 2011, 2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

#import "PropertiesUtilities.h"
#import "JNIUtilities.h"

@implementation PropertiesUtilities

+ (NSString *) javaSystemPropertyForKey:(NSString *)key withEnv:(JNIEnv *)env {
    DECLARE_CLASS_RETURN(jc_System, "java/lang/System", nil);
    DECLARE_STATIC_METHOD_RETURN(jm_getProperty, jc_System,
                                 "getProperty", "(Ljava/lang/String;)Ljava/lang/String;", nil);

    jstring jKey = NSStringToJavaString(env, key);
    jstring jValue = (*env)->CallStaticObjectMethod(env, jc_System, jm_getProperty, jKey);
    (*env)->DeleteLocalRef(env, jKey);
    CHECK_EXCEPTION_NULL_RETURN(jValue, nil);

    NSString *value = JavaStringToNSString(env, jValue);
    (*env)->DeleteLocalRef(env, jValue);
    return value;
}

@end
